package jumbo.xml;

import org.xml.sax.AttributeList;
import org.xml.sax.HandlerBase;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.Parser;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import org.xml.sax.helpers.ParserFactory;

import com.microstar.xml.XmlParser;

import java.awt.BorderLayout;

import java.lang.ClassNotFoundException;
import java.lang.Exception;
import java.lang.IllegalAccessException;
import java.lang.InstantiationException;
import java.lang.System;

import java.io.IOException;
import java.io.StringReader;

import java.net.MalformedURLException;
import java.net.URL;

import java.util.Hashtable;
import java.util.Stack;
import java.util.Vector;

import com.sun.java.swing.JFrame;
import com.sun.java.swing.JScrollPane;

import jumbo.xml.gui.XText;

/**
<P> David Megginson's SAX demo, amended by PeterMR to build a simple
jumbo.xml.XTree. Comments are from both authors. Where possible I have left in
David's comments in case others wish to hack this code. David's code
is commented out through slash-star-dash-dash.
</P>
@author David Megginson, 1998
@author Peter Murray-Rust, 1998
*/

public class SAX extends HandlerBase {

// parsers available (as class names). These are hardcoded, but we 
// should also allows write-ins in JComboBox

    public final static int AELFRED = 0;
	public final static int DXP		= AELFRED+1;
	public final static int LARK	= DXP+1;
	public final static int MSXML	= LARK+1;
	public final static int XP		= MSXML+1;

	public final static int nParserTypes = XP+1;

	public final static String[] parserTypeStrings = {
		"com.microstar.xml.SAXDriver",
		"com.datachannel.xml.sax.DXPDriver",
		"com.megginson.sax.LarkDriver",
		"com.megginson.sax.MSXMLDriver",
		"com.jclark.xml.sax.Driver",
	};
	public final static int defaultParserType = AELFRED;
	public final static String defaultParserClassName = parserTypeStrings[defaultParserType];

	XTree xTree = null;			// tree, if built
	XNode parentNode = null;    // current parent node while building
    Stack xNodeStack = new Stack();		// for holding higher level nodes

// archive various events
    Hashtable notationTable = null;
    Hashtable entityTable = null;
    Vector warningVector = new Vector();
    Vector errorVector = new Vector();
    Vector fatalErrorVector = new Vector();

// URL from SystemId
	String urlString = null;
// Reader from a string
	StringReader reader = null;

/** create a new SAXTree using the default parser
@param String urlString the URL with the XML document
@exception XException IO problems or parsing errors
*/
    public SAX(String urlString) throws XException {
		this(urlString, defaultParserClassName);
	}

/** create a new SAXTree.
@param String urlString the URL with the XML document
@param String parserClassName the parser to use (as a full class name)
@exception XException IO problems or parsing errors
*/
    public SAX(String urlString, String parserClassName) 
		throws XException {
		this(urlString, null, parserClassName);
	}

/** Create a SAX from a UTF-16 character stream. 
From DM's CharacterStreamDemo; the default parser is used. Precede
this by a call of the form:<BR>
<TT> StringReader reader = new StringReader(string); </TT>
@param StringReader reader the StringReader created from the String
@exception XException IO problems or parsing errors
*/
    public SAX(StringReader reader) throws XException {
		this(null, reader, defaultParserClassName);
	}

    private SAX(String urlString, StringReader reader, 
		String parserClassName) throws XException {
	    try {
	        Parser parser = makeParser(parserClassName);
            parser.setDocumentHandler(this);
            parser.setDTDHandler(this);
            parser.setErrorHandler(this);
            parser.setEntityResolver(this);
            if (urlString != null) {
				this.urlString = Util.makeAbsoluteURL(urlString);
				parser.parse(this.urlString);
			} else if (reader != null) {
				this.reader = reader;
				parser.parse(new InputSource(reader));
			}
	    } catch (SAXParseException spe) {
	        printErrors();
			displayError(spe);
	    } catch (SAXException se) {
		    Exception ee = se.getException();
		    if (ee == null) {
		    } else if (ee instanceof SAXParseException) {
				System.out.println("Fatal: "+getErrorMessage((SAXParseException) ee));
		    } else {
			    throw new XException("Fatal Exception: "+se.toString());
		    }
	    } catch (NullPointerException npe) {
  		    throw new XException(npe);
	    } catch (OutOfMemoryError oome) {
  		    throw new XException(""+oome);
	    } catch (Exception e) {
  		    throw new XException(e);
	    }
    }

  //////////////////////////////////////////////////////////////////////
  // Default implementation of the EntityResolver interface.
  //////////////////////////////////////////////////////////////////////


  /**
    * Resolve an external entity.
    * Always return null, so that the parser will use the system
    * identifier provided in the XML document.  This method implements
    * the SAX default behaviour: application writers can override it
    * in a subclass to do special translations such as catalog lookups
    * or URI redirection.
    * @param publicId The public identifer, or null if none is
    *                 available.
    * @param systemId The system identifier provided in the XML 
    *                 document.
    * @return The new input source, or null to require the
    *         default behaviour.
    * @see org.xml.sax.EntityResolver#resolveEntity
    * @exception SAXException IO problems or parsing errors
    */
    public InputSource resolveEntity (String publicId, String systemId)
        throws SAXException {
	    System.out.println("Entity: "+systemId);
	    if (entityTable == null) entityTable = new Hashtable();
   	    entityTable.put(systemId, systemId);
	    return null;
    }


  //////////////////////////////////////////////////////////////////////
  // Default implementation of DTDHandler interface.
  //////////////////////////////////////////////////////////////////////


  /**
    * Receive notification of a notation declaration.
    * By default, do nothing.  Application writers may override this
    * method in a subclass if they wish to keep track of the notations
    * declared in a document.
    * @param name The notation name.
    * @param publicId The notation public identifier, or null if not
    *                 available.
    * @param systemId The notation system identifier.
    * @see org.xml.sax.DTDHandler#notationDecl
    */
    public void notationDecl 
	    (String name, String publicId, String systemId) {
	    System.out.println("Notation: "+name+" "+publicId+" "+systemId);
	    if (notationTable == null) notationTable = new Hashtable();
	    notationTable.put(name, systemId);
    }


  /**
    * Receive notification of an unparsed entity declaration.
    * By default, do nothing.  Application writers may override this
    * method in a subclass to keep track of the unparsed entities
    * declared in a document.
    * @param name The entity name.
    * @param publicId The entity public identifier, or null if not
    *                 available.
    * @param systemId The entity system identifier.
    * @param notationName The name of the associated notation.
    * @see org.xml.sax.DTDHandler#unparsedEntityDecl
    */
    public void unparsedEntityDecl (String name, String publicId,
				  String systemId, String notationName) {
	    System.out.println("UnparsedEntity: "+name+" "+publicId+" "+systemId+" "+notationName);
    }

  //////////////////////////////////////////////////////////////////////
  // Default implementation of DocumentHandler interface.
  //////////////////////////////////////////////////////////////////////


  /**
    * Receive a Locator object for document events.
    * By default, do nothing.  Application writers may override this
    * method in a subclass if they wish to store the locator for use
    * with other document events.
    * @param locator A locator for all SAX document events.
    * @see org.xml.sax.DocumentHandler#setDocumentLocator
    * @see org.xml.sax.Locator
    */
    public void setDocumentLocator (Locator locator) {
//  	    System.out.println("Locator: "+locator);
    }


  //////////////////////////////////////////////////////////////////////
  // Override some SAXDocumentHandler methods from the base class.
  //////////////////////////////////////////////////////////////////////


  /**
    * Receive notification of the beginning of the document.
    * By default, do nothing.  Application writers may override this
    * method in a subclass to take specific actions at the beginning
    * of a document (such as allocating the root node of a tree or
    * creating an output file).
    * @see org.xml.sax.DocumentHandler#startDocument
    * @exception SAXException IO problems or parsing errors
    */
    public void startDocument () throws SAXException {
//        System.out.print("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\n");
		System.out.println("Start Document");
    }

  /**
    * Receive notification of the end of the document.
    * By default, do nothing.  Application writers may override this
    * method in a subclass to take specific actions at the beginning
    * of a document (such as finalising a tree or closing an output
    * file).
    * @see org.xml.sax.DocumentHandler#endDocument
    * @exception SAXException IO problems or parsing errors
    */
    public void endDocument() throws SAXException {
	    if (fatalErrorVector.size() > 0) {
		    throw new SAXException("Parse errors; invalid document");
	    }
    }



  /**
    * Receive notification of the start of an element.
    * By default, do nothing.  Application writers may override this
    * method in a subclass to take specific actions at the start of
    * each element (such as allocating a new tree node or writing
    * output to a file).
    * @param name The element type name.
    * @param attributes The specified or defaulted attributes.
    * @see org.xml.sax.DocumentHandler#startElement
    * @exception SAXException IO problems or parsing errors
    */
    public void startElement (String name, AttributeList attributes)
        throws SAXException {
/*-- DavidM
    System.out.print("<" + name);
--*/
        XNode currentNode = new XNode(name, name);
		if (parentNode == null) {
	        xTree = new XTree(currentNode);
            parentNode = currentNode;
		} else {
			parentNode.add(currentNode);
		}
        xNodeStack.push(parentNode);
	    parentNode = currentNode;

    	for (int i = 0; i < attributes.getLength(); i++) {
/*-- DavidM
      System.out.print(" " +
		       attributes.getName(i) +
		       "=\"" +
		       attributes.getValue(i) +
		       '"');
--*/
	    	try {
		    	currentNode.addAttribute(attributes.getName(i), attributes.getValue(i), Attribute.CDATA);
		    } catch (Exception e) {System.out.println("BUG: "+e);}
        }
/*-- DavidM
    System.out.print(">");
--*/

/*-- old SAX
    Enumeration attNames = atts.getAttributeNames();
    if (attNames == null) {
      System.out.println("[Attributes not available]");
    } else {
      while (attNames.hasMoreElements()) {
        String aname = (String)attNames.nextElement();
		int type = Attribute.CDATA;
      
        if (atts.isId(aname)) {
		    type = Attribute.ID;
        } else if (atts.isIdref(aname)) {
			type = Attribute.IDREF;
        } else if (atts.isEntity(aname)) {
			type = Attribute.NDATA;
        } else if (atts.isNotation(aname)) {
			type = Attribute.NOTATION;
        }

//      System.out.println("Attribute: " +
//			 aname +
//			 "=\"" +
//			 atts.getValue(aname) +
//			 '"' +
//			 special);
		try {
			currentNode.addAttribute(aname, atts.getValue(aname), type);
		} catch (Exception e) {System.out.println("BUG: "+e);}
	  }
---*/

    }


  /**
    * Receive notification of the end of an element.
    * By default, do nothing.  Application writers may override this
    * method in a subclass to take specific actions at the end of
    * each element (such as finalising a tree node or writing
    * output to a file).
    * @param name The element type name.
    * @param attributes The specified or defaulted attributes.
    * @see org.xml.sax.DocumentHandler#endElement
    * @exception SAXException IO problems or parsing errors
    */
  public void endElement(String name) throws SAXException {
      parentNode = (XNode) xNodeStack.pop();
/*--
    System.out.print("</" + name + ">");
--*/
    }

  /**
    * Receive notification of character data inside an element.
    * By default, do nothing.  Application writers may override this
    * method to take specific actions for each chunk of character data
    * (such as adding the data to a node or buffer, or printing it to
    * a file).
    * @param ch The characters.
    * @param start The start position in the character array.
    * @param length The number of characters to use from the
    *               character array.
    * @see org.xml.sax.DocumentHandler#characters
    * @exception SAXException IO problems or parsing errors
    */
    public void characters (char ch[], int start, int length) 
  	    throws SAXException {
/*--
    display(ch, start, length);
--*/
	    String s = new String(ch, 0, length);
        parentNode.add(new PCDATANode(s));
    }


  /**
    * Receive notification of ignorable whitespace in element content.
    * By default, do nothing.  Application writers may override this
    * method to take specific actions for each chunk of ignorable
    * whitespace (such as adding data to a node or buffer, or printing
    * it to a file).
    * @param ch The whitespace characters.
    * @param start The start position in the character array.
    * @param length The number of characters to use from the
    *               character array.
    * @see org.xml.sax.DocumentHandler#ignorableWhitespace
    * @exception SAXException IO problems or parsing errors
    */
    public void ignorableWhitespace (char ch[], int start, int length)
        throws SAXException {
/*--
    display(ch, start, length);
--*/
	    String s = new String(ch, 0, length);
	    PCDATANode pcdataNode = new PCDATANode(s);
	    pcdataNode.addAttribute("ignorable", "yes", Attribute.CDATA);
    }


  /**
    * Receive notification of a processing instruction.
    * By default, do nothing.  Application writers may override this
    * method in a subclass to take specific actions for each
    * processing instruction, such as setting status variables or
    * invoking other methods.
    * @param target The processing instruction target.
    * @param data The processing instruction data, or null if
    *             none is supplied.
    * @see org.xml.sax.DocumentHandler#processingInstruction
    * @exception SAXException IO problems or parsing errors
    */
    public void processingInstruction (String target, String data)
        throws SAXException {
/*--
    System.out.print("<?" + target);
    if (data != null) {
      System.out.print(" " + data);
    }
    System.out.print("?>");
--*/
		String xmlTarget = target.toLowerCase();
		if (parentNode == null) {
			if (xmlTarget.equals("xml:namespace")) {
				System.out.println("Namespace: "+data);
			} else {
				System.out.println("PI not stored: <?"+target+" "+data+"?>");
			}
		} else {
			if (xmlTarget.equals("xml:namespace")) {
				System.out.println("Namespace NOT allowed here: "+xmlTarget);
			} else {
				parentNode.add(new PINode(target, data));
			}
		} 
    }

  //////////////////////////////////////////////////////////////////////
  // Default implementation of the ErrorHandler interface.
  //////////////////////////////////////////////////////////////////////


  /**
    * Receive notification of a parser warning.
    * The default implementation does nothing.  Application writers
    * may override this method in a subclass to take specific actions
    * for each warning, such as inserting the message in a log file or
    * printing it to the console.
    * @param e The warning information encoded as an exception.
    * @see org.xml.sax.ErrorHandler#warning
    * @see org.xml.sax.SAXParseException
    * @exception SAXException IO problems or parsing errors
    */
    public void warning (SAXParseException e) throws SAXException {
		String message = "Warning: "+getErrorMessage(e);
	    warningVector.addElement(e);
	    System.out.println("*** "+message+" ***");
    // no op
    }


  /**
    * Receive notification of a recoverable parser error.
    * The default implementation does nothing.  Application writers
    * may override this method in a subclass to take specific actions
    * for each error, such as inserting the message in a log file or
    * printing it to the console.
    * @param e The warning information encoded as an exception.
    * @see org.xml.sax.ErrorHandler#warning
    * @see org.xml.sax.SAXParseException
    * @exception SAXException IO problems or parsing errors
    */
    public void error (SAXParseException e) throws SAXException {
		String message = "Error: "+getErrorMessage(e);
	    errorVector.addElement(e);
	    System.out.println("*** "+message+" ***");
    }


  /**
    * Report a fatal XML parsing error.
    * The default implementation throws a SAXParseException.
    * Application writers may override this method in a subclass if
    * they need to take specific actions for each fatal error (such as
    * collecting all of the errors into a single report): in any case,
    * the application must stop all regular processing when this
    * method is invoked, since the document is no longer reliable, and
    * the parser may no longer report non-error events.
    * @param e The error information encoded as an exception.
    * @see org.xml.sax.ErrorHandler#fatalError
    * @see org.xml.sax.SAXParseException
    * @exception SAXException IO problems or parsing errors
    */
    public void fatalError (SAXParseException e) throws SAXException {
		String message = "Fatal Error: "+getErrorMessage(e);
	    fatalErrorVector.addElement(e);
//	    System.out.println("*** "+message+" ***");
	    throw e;
    }

  //////////////////////////////////////////////////////////////////////
  // Utility routines.
  //////////////////////////////////////////////////////////////////////


  /**
    * Given a SAX parser class name, return an instance of the parser.
    */
    private static Parser makeParser (String className)
        throws ClassNotFoundException,
            InstantiationException,
            IllegalAccessException
    {
        Class c = Class.forName(className);
//		System.out.println("Making parser: "+className);
        return (Parser)(c.newInstance());
    }


  /**
    * If a URL is relative, make it absolute against the current directory.
    */
    private static String makeAbsoluteURL (String url)
        throws java.net.MalformedURLException {
        URL baseURL;

        String currentDirectory = System.getProperty("user.dir");
        String fileSep = System.getProperty("file.separator");
        String file = currentDirectory.replace(fileSep.charAt(0), '/') + '/';

        if (file.charAt(0) != '/') {
            file = "/" + file;
        }
        baseURL = new URL("file", null, file);

        return new URL(baseURL, url).toString();
    }


  /**
    * Display text, escaping some characters.
    */
    private static void display (char ch[], int start, int length) {
        for (int i = start; i < start + length; i++) {
            switch (ch[i]) {
                case '<':
 	                System.out.print("&lt;");
   	                break;
                case '&':
	                System.out.print("&amp;");
	                break;
                case '>':
	                System.out.print("&gt;");
	                break;
                default:
	                if ((ch[i] < ' ' || ch[i] > 'z') && ch[i] != '\n') {
	                    System.out.print("&#" + (int)ch[i] + ';');
	                } else {
	                    System.out.print(ch[i]);
		            }
	        }
        }
    }

/** get an error message string from a SAXParseException
*/
    public String getErrorMessage(SAXParseException e) {
	    String messageString = "";
	    if (e.getPublicId() != null) {
	        messageString += " PublicID: "+e.getPublicId()+";";
    	}
	    if (e.getSystemId() != null) {
		    messageString += " SystemID: "+e.getSystemId()+";";
    	}
	    if (e.getMessage() != null) {
		    messageString += " "+e.getMessage()+";";
    	}
        if (e.getLineNumber() > 0) {
		    messageString += " at Line: "+e.getLineNumber()+";";
	    }
        if (e.getColumnNumber() > 0) {
		    messageString += " at Col: "+e.getColumnNumber()+";";
	    }
	    return messageString;
    }
/** return the XTree created by SAX
@return XTree the XTree
*/
    public XTree getTree() {
        return xTree;
    }

/** return the entity table created by SAX (null if no entities)
@return Hashtable the hashtable of entities
*/
    public Hashtable getEntityTable() {
        return entityTable;
    }

/** return the notation table created by SAX (null if no notations)
@return Hashtable the hashtable of notations
*/
    public Hashtable getNotationTable() {
        return notationTable;
    }

/** print the accumulated errors
*/
    public void printErrors() {
	    System.out.println("---- Errors ----");
		printError("Warning: ", warningVector);
		printError("Error: ", errorVector);
		printError("Fatal Error: ", fatalErrorVector);
	    System.out.println("----------------");
    }

    private void printError(String type, Vector vector) {
		if (vector.size() > 0) {
		    for (int i = 0; i < vector.size(); i++) {
		  	    SAXParseException spe = (SAXParseException) vector.elementAt(i);
				System.out.println(type+getErrorMessage(spe));
		    }
	    }
    }
/** display the input and highlight the first fatal error
*/
    public void displayError(SAXParseException spe) {
		int line = spe.getLineNumber();
		int col = spe.getColumnNumber();
		Exception exc = spe.getException();
		String message = getErrorMessage(spe);

		JFrame jFrame = new JFrame("Input Errors");
		XText xText = null;
		boolean ok = true;
		try {
			if (urlString != null) {
				xText = new XText(new URL(urlString));
				xText.highlight(line, col);
			} else if (reader != null) {
				System.out.println("Reader errors not yet implemented");
			}
		} catch (MalformedURLException mfe) {
			Util.bug(mfe);
		} catch (IOException ioe) {
			ok = false;
		}
		if (!ok && exc != null) {
			xText.addText(exc.toString());
		}
// we'll give this a menu later
//		JPanel messageXText = new XText();
//		messageXText.addText(message);
//		jFrame.getContentPane().add
//			(messageXText.getJTextPane(), BorderLayout.NORTH);
		XText messageXText = new XText();
		messageXText.addText(message);
		jFrame.getContentPane().add
			(messageXText.getJTextPane(), BorderLayout.SOUTH);
		jFrame.getContentPane().add
			(new JScrollPane(xText.getJTextPane()), BorderLayout.CENTER);
		jFrame.pack();
		jFrame.setSize(300,400);
		jFrame.show();
    }
/**
  * Main entry point for an application.
  * @exception Exception almost anything in the galaxy could go wrong
  */
    public static void main (String args[])
        throws Exception {
        String parserClassName;
		XTree xTree = null;

        if (args.length != 1) {
            System.err.println("Usage: java Demo <document>");
            System.exit(2);
        }

        parserClassName = System.getProperty("org.xml.sax.parser",
					 defaultParserClassName);

	    try {
  	        SAX sax = new SAX(args[0], parserClassName);
	        xTree = sax.getTree();
// customise display
     	    XTreeOptions options = new XTreeOptions();
	        options.displayType = XTreeOptions.SPLITPANE;
	        xTree.setOptions(options);
	        xTree.display();
// make a simple tree from a string
			String test = "<greetings>Hello world<shriek/></greetings>";
			SAX sax1 = new SAX(new StringReader(test));
			XTree xTree1 = sax1.getTree();
			xTree1.setOptions(options);
			xTree1.display();
 	    } catch (XException e) {
			System.out.println("<!-- SAX Exception: ");
		    e.printStackTrace();
		    System.out.println(e.toString());
System.out.println(" --!>");
		    System.exit(1);
		}
    }

}