package jumbo.xml.gui;

import com.sun.java.swing.JScrollPane;
import com.sun.java.swing.JTabbedPane;

import java.awt.Component;

import java.util.Hashtable;

/** a simple implementation of JTabbedPane to support jumbo.xml. Components
are cached in a Hashtable so cannot be added several times by mistake.
Tooltips are enabled.
@author P.Murray-Rust, 1998
*/

public class TabbedPane extends JTabbedPane {
/** the initial maximum character length of tabs*/
	public static final int MAXTABLENGTH = 8;
	Hashtable tabTable = new Hashtable();
	int maxTabLength = MAXTABLENGTH;

/** default constructor */
	public TabbedPane() {
	}

/** add a component, supported by title (on the tab) and tooltip for that tab
@param Component component the component
@param String title the title (truncated if too long)
@param String toolTip the tooltip for the tab
*/
	public void add(Component component, String title, String toolTip) { 
		if (title.length() > maxTabLength) title = title.substring(0,maxTabLength);
		Component tabComponent = null;
// new component? if so, make and add
		if (tabTable.get(component) == null) {
			tabComponent = new JScrollPane(component);
			this.addTab(title, null, tabComponent, toolTip);
			tabTable.put(component, tabComponent);
		} else {
			tabComponent = (Component) tabTable.get(component);
		}
// and select it
		int idx = this.indexOfComponent(tabComponent);
		this.setSelectedIndex(idx);
	}

/** change the maximum tab length. Will not affect existing tabs
@param int max new maximum length of tab in characters.
*/
	public void setMaxTabLength(int max) {
		maxTabLength = max;
	}
}