package jumbo.xml.gui;

// hacked from SwingSet examples TableExample3
/**
 * An example showing the JTable with a dataModel that is not derived 
 * from a database. We add the optional TableSorter object to give the 
 * JTable the ability to sort.   
 *
 * @version 1.3 10/14/97
 * @author Philip Milne
 */

import com.sun.java.swing.DefaultCellEditor;
import com.sun.java.swing.ImageIcon;
import com.sun.java.swing.JButton;
import com.sun.java.swing.JComboBox;
import com.sun.java.swing.JComponent;
import com.sun.java.swing.JFrame;
import com.sun.java.swing.JLabel;
import com.sun.java.swing.JPanel;
import com.sun.java.swing.JTable;
import com.sun.java.swing.JScrollPane;
import com.sun.java.swing.SwingConstants;

import com.sun.java.swing.table.AbstractTableModel;
import com.sun.java.swing.table.DefaultTableCellRenderer;
import com.sun.java.swing.table.DefaultTableModel;
import com.sun.java.swing.table.JTableHeader;
import com.sun.java.swing.table.TableColumn;
import com.sun.java.swing.table.TableModel;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;

import java.net.URL;

import java.util.Vector;

/** a simple subset of JTable, using defaults. getDisplayComponent()
returns a Component for embedding elsewhere if required and takes
care of JTableHeader (I hope)
*/
public class XTable extends JTable {
	DefaultTableModel defaultTableModel;
	int nrows = 0;
	int ncols = 0;
	String name;
	JFrame jFrame = null;	// for a free-standing display

/** make the table from already constructed Object arrays [rows][cols]
and column names [cols]
@param name	name of the table
@param inData a [row][column] array of Objects
@param columnNames a [column] vector of Strings
@exception IllegalArgumentException lengths of arrays are inconsistent
*/
	public XTable(String name, Object[][] inData, String[] columnNames) {
		super(inData, columnNames);
		this.name = name;
		ncols = columnNames.length;
		nrows = inData[0].length;
//		dataModel = new XAbstractTableModel(columnNames, inData);
//		this.setModel(dataModel);
		defaultTableModel = (DefaultTableModel) this.getModel();
	}

/** add Component to a cell - see SwingSet example
*/
	public void addComponent(String columnName /*, JComponent component, int iRow*/) {

        TableColumn column = this.getColumn(columnName);
        
        DefaultTableCellRenderer columnRenderer = new DefaultTableCellRenderer() {
		    public void setValue(Object value) { 
		        int cellValue = (value instanceof Number) ? ((Number)value).intValue() : 0; 
			    setForeground((cellValue > 30) ? Color.black : Color.red); 
				setText((value == null) ? "" : value.toString());
		    } 
			public Component getTableCellRendererComponent
				(JTable table, Object value, boolean isSelected, boolean hasFocus,
				int row, int column) {
				if (false) {
					return null;
				} else if (value instanceof Integer) {
//					JButton jButton = new JButton("Button: "+value.toString());
//					jButton.setEnabled(true);
					ImageIcon imageIcon = new ImageIcon("blue-ball.gif");
					return new JLabel("L: "+value.toString(), imageIcon, SwingConstants.RIGHT);
				} else if (value != null) {
					return new JLabel("Label: "+value.toString());
				} else {
					return new JLabel("Null");
				}
			}
		};			
        columnRenderer.setToolTipText("Click for combo box"); 
        column.setCellRenderer(columnRenderer); 
	}

/** add ComboBox editing to a column - see SwingSet example
*/
	public void addComboBoxEditor(String columnName, String[] values) {

		JComboBox comboBox = new JComboBox();
		for (int i = 0; i < values.length; i++) {
			comboBox.addItem(values[i]);
		}
        TableColumn column = this.getColumn(columnName); 
// Use the combo box as the editor in the column. 
        column.setCellEditor(new DefaultCellEditor(comboBox)); 
        
        DefaultTableCellRenderer columnRenderer = new DefaultTableCellRenderer(); 
        columnRenderer.setToolTipText("Click for combo box"); 
        column.setCellRenderer(columnRenderer); 
        
/*--
        // Set a tooltip for the header of the colors column. 
	TableCellRenderer headerRenderer = colorColumn.getHeaderRenderer(); 
	if (headerRenderer instanceof DefaultTableCellRenderer)
	    ((DefaultTableCellRenderer)headerRenderer).setToolTipText("Hi Mom!");
--*/
	}

/** get the name of the table
@return String name the name of the table
*/	
	public String getName() {return name;}

/** delete a row
*/
	public void removeRow(int i) {
		defaultTableModel.removeRow(i);
	}

/** add a row at end
*/
	public void addRow(Object[] row) {
		defaultTableModel.addRow(row);
	}
															
/** get information
*/
	public int getRowCount() {
		return defaultTableModel.getRowCount();
	}

/** get Row as a Vector of objects (silly value of iRow returns null)
*/
	public Vector getRow(int iRow) {
		Vector v = defaultTableModel.getDataVector();
		Vector rowVector = null;
		try {
			rowVector = (Vector) v.elementAt(iRow);
		} catch (ArrayIndexOutOfBoundsException e) {}
		return rowVector;
	}

/** get Cell as an object (silly value of iRow, jCol returns null)
*/
	public Object getCell(int iRow, int jCol) {
		Vector rowVector = this.getRow(iRow);
		if (rowVector == null) return null;
		Object cell = null;
		try {
			cell =  rowVector.elementAt(jCol);
		} catch (ArrayIndexOutOfBoundsException e) {}
		return cell;
	}

/** returns a JPanel with a scrolling Table (still experimental)
@return JPanel JPanel containing a JScrollPane
*/
	public JPanel getDisplayComponent() {
		JPanel jPanel = new JPanel();
		jPanel.setLayout(new BorderLayout());
        JScrollPane scrollPane = JTable.createScrollPaneForTable(this/*tableView*/);
		jPanel.add(scrollPane, BorderLayout.CENTER);
		JTableHeader jTableHeader = this.createDefaultTableHeader();
//		if (jTableHeader != null) {
//			jPanel.add(jTableHeader, BorderLayout.SOUTH);
//		}
		return jPanel;
	}

/** a default display in a JFrame
*/
	public void display() {
        jFrame = new JFrame(name);
        jFrame.getContentPane().add(getDisplayComponent());
        jFrame.pack();
        jFrame.setVisible(true);
    }

	public JFrame getJFrame() {return jFrame;}

/** test routine */
	public static void main(String[] args) {
		if (args.length == 0) {
			System.out.println("Usage: XTable foo");
			System.exit(1);
		}
		String columnNames[] = {"Col1", "Col2", "Col3", "Col4"};
		Object data[][] = {
			{"Row1", new Integer(12), Color.red, "one"},
			{"Row2", new Integer(22), Color.green, "two"},
			{"Row3", "Fred", Color.blue, "three"},
			{"Row4", new Integer(42), Color.pink, "one"},
			{"Row5", new Integer(52), Color.red, "two"},
			{"Row6", new Integer(62), Color.green, "three"},
			{"Row7", new Integer(72), Color.blue, "four"},
			{"Row8", new Integer(82), Color.pink, "one"},
			{"Row9", new Integer(92), Color.red, "two"},
			{"Row10", Color.pink, Color.green, "three"},
			{"Row11", new Integer(112), Color.blue, "two"},
			{"Row12", new Integer(122), Color.pink, "one"},
		};
		try {
			XTable xTable = new XTable("Table", data, columnNames);
			String[] values = {"one", "two", "three"};
			xTable.addComboBoxEditor("Col4", values);
			xTable.addComponent("Col2");
	        JFrame jFrame = new JFrame("Table");
			JPanel jPanel = new JPanel();
			jPanel.setLayout(new BorderLayout());
			jPanel.add(xTable.getDisplayComponent(), BorderLayout.CENTER);
			JButton jButton = new JButton("Delete Selected Row");
			jPanel.add(jButton, BorderLayout.NORTH);
			final XTable x1Table = xTable;
			ActionListener al = new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					int selectedRow = x1Table.getSelectedRow();
					if (selectedRow == -1) {
						System.out.println("No row selected");
					} else {
						x1Table.removeRow(selectedRow);
					}
				}
			};
			jButton.addActionListener(al);
			jButton = new JButton("Add Row at end");
			jPanel.add(jButton, BorderLayout.SOUTH);
			al = new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					Object[] row = {"Row99", new Integer(99), Color.pink, "ninetynine"};
					x1Table.addRow(row);
				}
			};
			jButton.addActionListener(al);
			jButton = new JButton("Print");
			jPanel.add(jButton, BorderLayout.WEST);
			al = new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					int nRows = x1Table.getRowCount();
					System.out.println("Content ("+nRows+")");
					for (int iRow = 0; iRow < nRows; iRow++) {
						Vector row = x1Table.getRow(iRow);
						for (int jCol = 0; jCol < row.size(); jCol++) {
							System.out.print(row.elementAt(jCol)+" : ");
						}
						System.out.println();
					}
				}
			};
			jButton.addActionListener(al);
			jFrame.getContentPane().add(jPanel);
			jFrame.pack();
			jFrame.setVisible(true);

//			xTable.display();
//			xTable.removeRow(6);  // removes row7?
		} catch (Exception e) {e.printStackTrace();}
	}
}
// this may be unnecessary
class XAbstractTableModel extends AbstractTableModel {
	String[] columnNames;
	Object[][] inData;
	public XAbstractTableModel(String[] columnNames, Object[][] inData) {
	//	super();
		this.columnNames = columnNames;
		this.inData = inData;
	}
// These methods always need to be implemented. 
    public int getColumnCount() {
		return columnNames.length; 
	} 
    public int getRowCount() {
		return inData.length;
	}
    public Object getValueAt(int row, int col) {
		return inData[row][col];
	}

// The default implementations of these methods in 
// AbstractTableModel would work, but we can refine them. 
//            public String getColumnName(int column) {return columnNames[column];}
//            public Class getColumnClass(int col) {return getValueAt(0,col).getClass();}
//            public boolean isCellEditable(int row, int col) {return true;} 
//            public void setValueAt(Object aValue, int row, int column) {
//                inData[row][column] = aValue; 
}
